@@ -182,8 +182,10 @@ struct supplicant_task {
 	char *netpath;
 	gboolean created;
 	enum supplicant_state state;
-	gboolean noscan;
+	gboolean scanning;
 	GSList *scan_results;
+	DBusPendingCall *scan_call;
+	DBusPendingCall *result_call;
 	struct iw_range *range;
 	gboolean disconnecting;
 };
@@ -329,6 +331,8 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 	return;
 
 failed:
+	dbus_message_unref(reply);
+
 	task_list = g_slist_remove(task_list, task);
 
 	connman_device_unref(task->device);
@@ -901,13 +905,41 @@ static int set_network(struct supplicant_task *task,
 	return 0;
 }
 
+static void scan_reply(DBusPendingCall *call, void *user_data)
+{
+	struct supplicant_task *task = user_data;
+	DBusMessage *reply;
+
+	DBG("task %p", task);
+
+	task->scan_call = NULL;
+
+	reply = dbus_pending_call_steal_reply(call);
+	if (reply == NULL)
+		return;
+
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
+		connman_device_set_scanning(task->device, FALSE);
+		goto done;
+	}
+
+	if (task->scanning == TRUE)
+		connman_device_set_scanning(task->device, TRUE);
+
+done:
+	dbus_message_unref(reply);
+}
+
+
 static int initiate_scan(struct supplicant_task *task)
 {
 	DBusMessage *message;
-	DBusPendingCall *call;
 
 	DBG("task %p", task);
 
+	if (task->scan_call != NULL)
+		return -EALREADY;
+
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
 					SUPPLICANT_INTF ".Interface", "scan");
 	if (message == NULL)
@@ -916,15 +948,23 @@ static int initiate_scan(struct supplicant_task *task)
 	dbus_message_set_auto_start(message, FALSE);
 
 	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
+					&task->scan_call, TIMEOUT) == FALSE) {
 		connman_error("Failed to initiate scan");
 		dbus_message_unref(message);
 		return -EIO;
 	}
 
+	if (task->scan_call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_pending_call_set_notify(task->scan_call, scan_reply, task, NULL);
+
 	dbus_message_unref(message);
 
-	return 0;
+	return -EINPROGRESS;
 }
 
 static struct {
@@ -1334,7 +1374,6 @@ done:
 static void get_properties(struct supplicant_task *task)
 {
 	DBusMessage *message;
-	DBusPendingCall *call;
 	char *path;
 
 	path = g_slist_nth_data(task->scan_results, 0);
@@ -1354,27 +1393,32 @@ static void get_properties(struct supplicant_task *task)
 	dbus_message_set_auto_start(message, FALSE);
 
 	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
+				&task->result_call, TIMEOUT) == FALSE) {
 		connman_error("Failed to get network properties");
 		dbus_message_unref(message);
 		goto noscan;
 	}
 
-	if (call == NULL) {
+	if (task->result_call == NULL) {
 		connman_error("D-Bus connection not available");
 		dbus_message_unref(message);
 		goto noscan;
 	}
 
-	dbus_pending_call_set_notify(call, properties_reply, task, NULL);
+	dbus_pending_call_set_notify(task->result_call,
+					properties_reply, task, NULL);
 
 	dbus_message_unref(message);
 
 	return;
 
 noscan:
-	if (task->noscan == FALSE)
+	task->result_call = NULL;
+
+	if (task->scanning == TRUE) {
 		connman_device_set_scanning(task->device, FALSE);
+		task->scanning = FALSE;
+	}
 }
 
 static void scan_results_reply(DBusPendingCall *call, void *user_data)
@@ -1431,17 +1475,23 @@ done:
 	dbus_message_unref(reply);
 
 noscan:
-	if (task->noscan == FALSE)
+	task->result_call = NULL;
+
+	if (task->scanning == TRUE) {
 		connman_device_set_scanning(task->device, FALSE);
+		task->scanning = FALSE;
+	}
 }
 
 static void scan_results_available(struct supplicant_task *task)
 {
 	DBusMessage *message;
-	DBusPendingCall *call;
 
 	DBG("task %p", task);
 
+	if (task->result_call != NULL)
+		return;
+
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
 						SUPPLICANT_INTF ".Interface",
 							"scanResults");
@@ -1451,20 +1501,21 @@ static void scan_results_available(struct supplicant_task *task)
 	dbus_message_set_auto_start(message, FALSE);
 
 	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
+				&task->result_call, TIMEOUT) == FALSE) {
 		connman_error("Failed to request scan result");
 		goto done;
 	}
 
-	if (task->noscan == FALSE)
-		connman_device_set_scanning(task->device, TRUE);
-
-	if (call == NULL) {
+	if (task->result_call == NULL) {
 		connman_error("D-Bus connection not available");
 		goto done;
 	}
 
-	dbus_pending_call_set_notify(call, scan_results_reply, task, NULL);
+	if (task->scanning == TRUE)
+		connman_device_set_scanning(task->device, TRUE);
+
+	dbus_pending_call_set_notify(task->result_call,
+					scan_results_reply, task, NULL);
 
 done:
 	dbus_message_unref(message);
@@ -1497,6 +1548,7 @@ static int task_connect(struct supplicant_task *task)
 	const char *address, *security, *passphrase;
 	const void *ssid;
 	unsigned int ssid_len;
+	int err;
 
 	address = connman_network_get_string(task->network, "Address");
 	security = connman_network_get_string(task->network, "WiFi.Security");
@@ -1520,7 +1572,9 @@ static int task_connect(struct supplicant_task *task)
 
 	set_network(task, ssid, ssid_len, address, security, passphrase);
 
-	enable_network(task);
+	err = enable_network(task);
+	if (err < 0)
+		return err;
 
 	return -EINPROGRESS;
 }
@@ -1552,31 +1606,13 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 	if (state == WPA_INVALID)
 		return;
 
-	task->state = state;
-
-	switch (task->state) {
-	case WPA_SCANNING:
-		task->noscan = TRUE;
-		connman_device_set_scanning(task->device, TRUE);
-		break;
-	case WPA_ASSOCIATING:
-	case WPA_ASSOCIATED:
-	case WPA_4WAY_HANDSHAKE:
-	case WPA_GROUP_HANDSHAKE:
-		task->noscan = TRUE;
-		break;
-	case WPA_COMPLETED:
-	case WPA_DISCONNECTED:
-		task->noscan = FALSE;
-		break;
-	case WPA_INACTIVE:
-		task->noscan = FALSE;
+	if (task->scanning == TRUE && state != WPA_SCANNING) {
 		connman_device_set_scanning(task->device, FALSE);
-		break;
-	case WPA_INVALID:
-		break;
+		task->scanning = FALSE;
 	}
 
+	task->state = state;
+
 	if (task->network == NULL)
 		return;
 
@@ -1588,14 +1624,15 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 
 		/* carrier on */
 		connman_network_set_connected(task->network, TRUE);
-		connman_device_set_scanning(task->device, FALSE);
 		break;
 
 	case WPA_DISCONNECTED:
 		disable_network(task);
 
+		/* carrier off */
+		connman_network_set_connected(task->network, FALSE);
+
 		if (task->disconnecting == TRUE) {
-			connman_network_set_connected(task->network, FALSE);
 			connman_network_unref(task->network);
 			task->disconnecting = FALSE;
 
@@ -1604,10 +1641,6 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 				task->pending_network = NULL;
 				task_connect(task);
 			}
-		} else {
-			/* carrier off */
-			connman_network_set_connected(task->network, FALSE);
-			connman_device_set_scanning(task->device, FALSE);
 		}
 		break;
 
@@ -1616,10 +1649,9 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		break;
 
 	case WPA_INACTIVE:
-		remove_network(task);
+		connman_network_set_connected(task->network, FALSE);
 
 		if (task->disconnecting == TRUE) {
-			connman_network_set_connected(task->network, FALSE);
 			connman_network_unref(task->network);
 			task->disconnecting = FALSE;
 
@@ -1701,7 +1733,7 @@ int supplicant_start(struct connman_device *device)
 	task->device = connman_device_ref(device);
 
 	task->created = FALSE;
-	task->noscan = FALSE;
+	task->scanning = FALSE;
 	task->state = WPA_INVALID;
 	task->disconnecting = FALSE;
 	task->pending_network = NULL;
@@ -1733,6 +1765,19 @@ int supplicant_stop(struct connman_device *device)
 
 	task_list = g_slist_remove(task_list, task);
 
+	if (task->scan_call != NULL) {
+		dbus_pending_call_cancel(task->scan_call);
+		task->scan_call = NULL;
+	}
+
+	if (task->result_call != NULL) {
+		dbus_pending_call_cancel(task->result_call);
+		task->result_call = NULL;
+	}
+
+	if (task->scanning == TRUE)
+		connman_device_set_scanning(task->device, FALSE);
+
 	disable_network(task);
 
 	remove_network(task);
@@ -1764,7 +1809,18 @@ int supplicant_scan(struct connman_device *device)
 		break;
 	}
 
+	task->scanning = TRUE;
+
 	err = initiate_scan(task);
+	if (err < 0) {
+		if (err == -EINPROGRESS)
+			return 0;
+
+		task->scanning = FALSE;
+		return err;
+	}
+
+	connman_device_set_scanning(task->device, TRUE);
 
 	return 0;
 }
