@@ -90,6 +90,17 @@ void connman_debug(const char *format, ...)
 	va_end(ap);
 }
 
+void __connman_toggle_debug(void)
+{
+	if (debug_enabled == TRUE) {
+		connman_info("Disabling debug output");
+		debug_enabled = FALSE;
+	} else {
+		connman_info("Enabling debug output");
+		debug_enabled = TRUE;
+	}
+}
+
 int __connman_log_init(gboolean detach, gboolean debug)
 {
 	int option = LOG_NDELAY | LOG_PID;
@@ -101,7 +112,8 @@ int __connman_log_init(gboolean detach, gboolean debug)
 
 	syslog(LOG_INFO, "Connection Manager version %s", VERSION);
 
-	debug_enabled = debug;
+	if (debug == TRUE)
+		__connman_toggle_debug();
 
 	return 0;
 }
@@ -112,3 +124,8 @@ void __connman_log_cleanup(void)
 
 	closelog();
 }
+
+gboolean __connman_debug_enabled(void)
+{
+	return debug_enabled;
+}
