@@ -64,7 +64,7 @@ static void append_path(gpointer value, gpointer user_data)
 	struct connman_service *service = value;
 	DBusMessageIter *iter = user_data;
 
-	if (service->path == NULL)
+	if (service->path == NULL || service->hidden == TRUE)
 		return;
 
 	dbus_message_iter_append_basic(iter, DBUS_TYPE_OBJECT_PATH,
@@ -950,39 +950,6 @@ int connman_service_set_favorite(struct connman_service *service,
 	return 0;
 }
 
-int __connman_service_set_carrier(struct connman_service *service,
-						connman_bool_t carrier)
-{
-	DBG("service %p carrier %d", service, carrier);
-
-	if (service == NULL)
-		return -EINVAL;
-
-	switch (service->type) {
-	case CONNMAN_SERVICE_TYPE_UNKNOWN:
-	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
-		return -EINVAL;
-	case CONNMAN_SERVICE_TYPE_ETHERNET:
-		break;
-	}
-
-	if (carrier == FALSE) {
-		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
-		state_changed(service);
-
-		service->state = CONNMAN_SERVICE_STATE_IDLE;
-		state_changed(service);
-	} else {
-		service->state = CONNMAN_SERVICE_STATE_CARRIER;
-		state_changed(service);
-	}
-
-	return connman_service_set_favorite(service, carrier);
-}
-
 static void default_changed(void)
 {
 	DBusMessage *signal;
@@ -1017,15 +984,9 @@ int __connman_service_indicate_state(struct connman_service *service,
 	if (service == NULL)
 		return -EINVAL;
 
-	if (state == CONNMAN_SERVICE_STATE_CARRIER)
-		return __connman_service_set_carrier(service, TRUE);
-
 	if (service->state == state)
 		return -EALREADY;
 
-	if (service->state == CONNMAN_SERVICE_STATE_READY)
-		default_changed();
-
 	if (service->state == CONNMAN_SERVICE_STATE_FAILURE &&
 				state == CONNMAN_SERVICE_STATE_IDLE)
 		return -EINVAL;
@@ -1052,6 +1013,14 @@ int __connman_service_indicate_state(struct connman_service *service,
 
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
+
+		__connman_notifier_connect(service->type);
+
+		default_changed();
+	} else if (state == CONNMAN_SERVICE_STATE_DISCONNECT) {
+		default_changed();
+
+		__connman_notifier_disconnect(service->type);
 	}
 
 	if (state == CONNMAN_SERVICE_STATE_FAILURE) {
@@ -1098,6 +1067,36 @@ int __connman_service_indicate_default(struct connman_service *service)
 	return 0;
 }
 
+static connman_bool_t prepare_network(struct connman_service *service)
+{
+	enum connman_network_type type;
+	unsigned int ssid_len;
+
+	type = connman_network_get_type(service->network);
+
+	switch (type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+		return FALSE;
+	case CONNMAN_NETWORK_TYPE_WIFI:
+		if (connman_network_get_blob(service->network, "WiFi.SSID",
+							&ssid_len) == NULL)
+			return FALSE;
+
+		connman_network_set_string(service->network,
+				"WiFi.Passphrase", service->passphrase);
+		break;
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		break;
+	}
+
+	return TRUE;
+}
+
 int __connman_service_connect(struct connman_service *service)
 {
 	int err;
@@ -1111,15 +1110,9 @@ int __connman_service_connect(struct connman_service *service)
 		return -EALREADY;
 
 	if (service->network != NULL) {
-		unsigned int ssid_len;
-
-		if (connman_network_get_blob(service->network, "WiFi.SSID",
-						     &ssid_len) == NULL)
+		if (prepare_network(service) == FALSE)
 			return -EINVAL;
 
-		connman_network_set_string(service->network,
-				"WiFi.Passphrase", service->passphrase);
-
 		err = __connman_network_connect(service->network);
 	} else if (service->device != NULL) {
 		if (service->favorite == FALSE)
@@ -1301,7 +1294,7 @@ int __connman_service_create_and_connect(DBusMessage *msg)
 	if (ssid_len < 1)
 		return -EINVAL;
 
-	device = __connman_element_find_device(CONNMAN_DEVICE_TYPE_WIFI);
+	device = __connman_element_find_device(CONNMAN_SERVICE_TYPE_WIFI);
 	if (device == NULL)
 		return -EOPNOTSUPP;
 
@@ -1509,12 +1502,21 @@ struct connman_service *__connman_service_create_from_device(struct connman_devi
 void __connman_service_remove_from_device(struct connman_device *device)
 {
 	struct connman_service *service;
+	enum connman_service_type type;
 
 	service = __connman_service_lookup_from_device(device);
 	if (service == NULL)
 		return;
 
+	type = service->type;
+
 	__connman_service_put(service);
+
+	default_changed();
+
+	__connman_notifier_disconnect(type);
+
+	__connman_service_auto_connect();
 }
 
 /**
